import collections
import re
import sqlite3

import sqliteparser

from ._exception import ISQLiteError

CURRENT_TIMESTAMP = "STRFTIME('%Y-%m-%d %H:%M:%f000+00:00', 'now')"


class Database:
    def __init__(self, connection_or_path, *, readonly=None, uri=False):
        # Validate arguments.
        if readonly is not None:
            if uri is True:
                raise ISQLiteError(
                    "The `readonly` parameter cannot be set if `uri` is True. Append "
                    + "'?mode=ro' (or omit it if you don't want your connection to be "
                    + "read-only) to your URI instead."
                )

            if not isinstance(connection_or_path, str):
                if readonly is not None:
                    raise ISQLiteError(
                        "The `readonly` parameter can only be set if a database path "
                        + "is passed, not an existing database connection."
                    )
        else:
            # Default value of `readonly` if not specified is False.
            readonly = False

        if isinstance(connection_or_path, str):
            if uri:
                path = connection_or_path
            else:
                if readonly is True:
                    path = f"file:{connection_or_path}?mode=ro"
                else:
                    path = f"file:{connection_or_path}"

            self.connection = sqlite3.connect(
                path, detect_types=sqlite3.PARSE_DECLTYPES, uri=True
            )
        else:
            self.connection = connection_or_path

        self.connection.row_factory = ordered_dict_row_factory
        self.connection.execute("PRAGMA foreign_keys = 1")
        self.cursor = self.connection.cursor()

    def get(self, table, *, where="1", values={}, camel_case=False):
        return self.sql(
            f"SELECT * FROM {table} WHERE {where}",
            values,
            camel_case=camel_case,
            multiple=False,
        )

    def get_by_rowid(self, table, rowid, **kwargs):
        return self.get(
            table, where="rowid = :rowid", values={"rowid": rowid}, **kwargs
        )

    def get_or_create(self, table, data, *, camel_case=False):
        if not data:
            raise ISQLiteError(
                "The `data` parameter to `get_or_create` cannot be empty."
            )

        query = " AND ".join(f"{key} = :{key}" for key in data)
        row = self.get(table, where=query, values=data)
        if row is None:
            pk = self.create(table, data)
            return self.get_by_rowid(table, pk, camel_case=camel_case)
        else:
            return row

    def list(
        self,
        table,
        *,
        where="1",
        values={},
        camel_case=False,
        limit=None,
        order_by=None,
        descending=None,
    ):
        # TODO(2021-04-30): Use `limit` parameter.
        if order_by is not None:
            if isinstance(order_by, (tuple, list)):
                order_by = ", ".join(order_by)

            direction = "DESC" if descending is True else "ASC"
            order_clause = f"ORDER BY {order_by} {direction}"
        else:
            if descending is not None:
                raise ISQLiteError(
                    "The `descending` parameter to `list` requires the `order_by` "
                    + "parameter to be set."
                )
            order_clause = ""

        return self.sql(
            f"SELECT * FROM {table} WHERE {where} {order_clause}",
            values,
            camel_case=camel_case,
        )

    def count(self, table, *, where="1", values={}):
        result = self.sql(
            f"SELECT COUNT(*) FROM {table} WHERE {where}",
            values,
            as_tuple=True,
            multiple=False,
        )
        return result[0]

    def create(self, table, data, *, auto_created_at=None, auto_last_updated_at=None):
        keys = list(data.keys())
        placeholders = ",".join("?" for _ in range(len(keys)))
        values = list(data.values())

        extra = []
        if auto_created_at and auto_created_at not in data:
            keys.append(auto_created_at)
            extra.append(CURRENT_TIMESTAMP)
        if auto_last_updated_at and auto_last_updated_at not in data:
            keys.append(auto_last_updated_at)
            extra.append(CURRENT_TIMESTAMP)

        if extra:
            extra = (", " if data else "") + ", ".join(extra)
        else:
            extra = ""

        self.cursor.execute(
            f"""
            INSERT INTO {table}({', '.join(keys)}) VALUES ({placeholders}{extra});
            """,
            values,
        )
        return self.cursor.lastrowid

    def create_many(
        self, table, data, *, auto_created_at=None, auto_last_updated_at=None
    ):
        if not data:
            return

        keys = list(data[0].keys())
        placeholders = ",".join("?" for _ in range(len(keys)))

        extra = []
        if auto_created_at and auto_created_at not in data[0]:
            keys.append(auto_created_at)
            extra.append(CURRENT_TIMESTAMP)
        if auto_last_updated_at and auto_last_updated_at not in data[0]:
            keys.append(auto_last_updated_at)
            extra.append(CURRENT_TIMESTAMP)

        if extra:
            extra = (", " if data else "") + ", ".join(extra)
        else:
            extra = ""

        self.cursor.executemany(
            f"""
            INSERT INTO {table}({', '.join(keys)}) VALUES ({placeholders}{extra});
            """,
            [tuple(d.values()) for d in data],
        )

    def update(self, table, data, *, where=None, values={}, auto_last_updated_at=None):
        updates = []
        for key, value in data.items():
            if key == "last_updated_at":
                continue

            placeholder = f"v{len(values)}"
            values[placeholder] = value
            updates.append(f"{key} = :{placeholder}")

        if auto_last_updated_at:
            updates.append(f"{auto_last_updated_at} = {CURRENT_TIMESTAMP}")

        updates = ", ".join(updates)
        where_clause = f"WHERE {where}" if where else ""
        self.cursor.execute(f"UPDATE {table} SET {updates} {where_clause}", values)

    def update_by_rowid(self, table, rowid, data):
        return self.update(table, data, where="rowid = :rowid", values={"rowid": rowid})

    def delete(self, table, *, where, values={}):
        self.sql(f"DELETE FROM {table} WHERE {where}", values=values)

    def delete_by_rowid(self, table, rowid):
        return self.delete(table, where="rowid = :rowid", values={"rowid": rowid})

    def sql(self, query, values={}, *, as_tuple=False, camel_case=False, multiple=True):
        if multiple:
            self.cursor.execute(query, values)
            rows = self.cursor.fetchall()
            if as_tuple:
                return [tuple(row.values()) for row in rows]

            if camel_case:
                return [row_to_camel_case(row) for row in rows]

            return rows
        else:
            self.cursor.execute(query + " LIMIT 1", values)
            row = self.cursor.fetchone()
            if row is None:
                return row

            if as_tuple:
                return tuple(row.values())

            if camel_case:
                return row_to_camel_case(row)

            return row

    def commit(self):
        self.connection.commit()

    def rollback(self):
        self.connection.rollback()

    def close(self):
        self.connection.commit()
        self.connection.close()

    def create_table(self, table_name, *columns, values={}):
        self.sql(f"CREATE TABLE {table_name}({','.join(columns)})", values)

    def drop_table(self, table_name):
        self.sql(f"DROP TABLE {table_name}")

    def add_column(self, table_name, column_name, column_def):
        self.sql(f"ALTER TABLE {table_name} ADD COLUMN {column_name} {column_def}")

    def drop_column(self, table_name, column_name):
        # ALTER TABLE ... DROP COLUMN is only supported since SQLite version 3.35, so we
        # implement it by hand here.
        create_table_statement = self._get_create_table_statement(table_name)
        new_columns = [
            column
            for column in create_table_statement.columns
            if column.name != column_name
        ]
        if len(new_columns) == len(create_table_statement.columns):
            # TODO(2021-05-25): Test this.
            raise ISQLiteError(f"{column_name} is not a column of {table_name}.")

        create_table_statement.columns = new_columns
        select = ", ".join(c.name for c in create_table_statement.columns)
        self._migrate_table(create_table_statement, select=select)

    def reorder_columns(self, table_name, column_names):
        create_table_statement = self._get_create_table_statement(table_name)
        column_map = collections.OrderedDict(
            (c.name, c) for c in create_table_statement.columns
        )
        if set(column_names) != set(column_map.keys()):
            raise ISQLiteError(
                "The set of reordered columns is not the same as the set of original "
                + "columns."
            )
        create_table_statement.columns = [column_map[name] for name in column_names]
        self._migrate_table(create_table_statement, select=", ".join(column_names))

    def alter_column(self, table_name, column_name, new_column):
        create_table_statement = self._get_create_table_statement(table_name)
        altered = False
        for i, column in enumerate(create_table_statement.columns):
            if column.name == column_name:
                # TODO(2021-05-25): Clean up this hacky logic. A better way would be to
                # call sqliteparser to parse `new_column` into a real `ast.Column`
                # object. But currently sqliteparser does not have a public API method
                # to do so.
                create_table_statement.columns[i] = sqliteparser.ast.Column(
                    column_name, constraints=[new_column]
                )
                altered = True
                break

        if not altered:
            # TODO(2021-05-25): Test this.
            raise ISQLiteError(f"{column_name} is not a column of {table_name}.")

        self._migrate_table(
            create_table_statement,
            select=", ".join(c.name for c in create_table_statement.columns),
        )

    def rename_column(self, table_name, old_column_name, new_column_name):
        # ALTER TABLE ... RENAME COLUMN is only supported since SQLite version 3.25, so
        # we implement it by hand here.
        raise NotImplementedError

    def _migrate_table(self, new_create_table_statement, *, select):
        name = new_create_table_statement.name
        # This procedure is copied from https://sqlite.org/lang_altertable.html
        try:
            self.sql("PRAGMA foreign_keys = 0")

            # Create the new table under a temporary name.
            tmp_table_name = f"isqlite_tmp_{name}"
            new_create_table_statement.name = tmp_table_name
            self.sql(str(new_create_table_statement))

            # Copy over all data from the old table into the new table using the
            # provided SELECT values.
            self.sql(f"INSERT INTO {tmp_table_name} SELECT {select} FROM {name}")

            # Drop the old table.
            self.sql(f"DROP TABLE {name}")

            # Rename the new table to the original name.
            self.sql(f"ALTER TABLE {tmp_table_name} RENAME TO {name}")

            # Check that no foreign key constraints have been violated.
            self.sql("PRAGMA foreign_key_check")
        except Exception as e:
            self.rollback()
            raise e
        else:
            self.commit()
        finally:
            self.sql("PRAGMA foreign_keys = 1")

    def _get_create_table_statement(self, table_name):
        sql = self.sql(
            "SELECT sql FROM sqlite_master WHERE type = 'table' AND name = :table",
            {"table": table_name},
            as_tuple=True,
            multiple=False,
        )[0]

        if not sql.endswith(";"):
            sql = sql + ";"

        return sqliteparser.parse(sql)[0]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close()


def ordered_dict_row_factory(cursor, row):
    return collections.OrderedDict(
        (column[0], row[i]) for i, column in enumerate(cursor.description)
    )


def row_to_camel_case(row):
    return collections.OrderedDict(
        (string_to_camel_case(key), value) for key, value in row.items()
    )


camel_case_pattern = re.compile(r"._([a-z])")


def string_to_camel_case(s):
    return camel_case_pattern.sub(lambda m: m.group(0)[0] + m.group(1).upper(), s)



import datetime
import decimal
import sqlite3

from ._core import Database
from ._exception import ISQLiteError


def sqlite3_convert_boolean(b):
    return b != b"0"


def sqlite3_convert_decimal(b):
    return decimal.Decimal(b.decode("utf8"))


def sqlite3_adapt_decimal(d):
    return str(d)


def sqlite3_convert_time(b):
    hour, minute, second = b.decode("utf8").split(":")
    return datetime.time(int(hour), int(minute), int(second))


def sqlite3_adapt_time(t):
    return str(t)


sqlite3.register_converter("BOOLEAN", sqlite3_convert_boolean)
sqlite3.register_converter("DECIMAL", sqlite3_convert_decimal)
sqlite3.register_adapter(decimal.Decimal, sqlite3_adapt_decimal)
sqlite3.register_converter("TIME", sqlite3_convert_time)
sqlite3.register_adapter(datetime.time, sqlite3_adapt_time)



import collections
import re
import sys

import click
from xcli import Table, colors, input2

from . import Database


@click.group()
def cli():
    pass


@cli.command(name="create")
@click.argument("path_to_database")
@click.argument("table")
def main_create(path_to_database, table):
    """
    Create a new row interactively.
    """
    with Database(path_to_database) as db:
        schema_row = db.sql(
            "SELECT sql FROM sqlite_master WHERE name = :name AND type = 'table'",
            {"name": table},
            multiple=False,
        )

        if schema_row is None:
            print(f"Table {table!r} not found.")
            sys.exit(1)

        schema = parse_create_table_statement(schema_row["sql"])

        payload = {}
        for name, definition in schema.items():
            if name in {"id", "created_at", "last_updated_at"}:
                continue

            column_type = definition.split(maxsplit=1)[0]
            print(colors.blue(name), definition)
            v = input2("? ", verify=lambda v: validate_column(column_type, v))

            if not v and column_type != "TEXT":
                v = None

            payload[name] = v

        pk = db.create(table, payload)
        print(f"Row {pk} created.")


@cli.command(name="delete")
@click.argument("path_to_database")
@click.argument("table")
@click.argument("pk", type=int)
def main_delete(path_to_database, table, pk):
    """
    Delete a row.
    """
    with Database(path_to_database) as db:
        try:
            row = db.get_by_rowid(table, pk)
        except Exception:
            # This can happen, e.g., if a timestamp column is invalidly formatted and
            # the Python wrapper around sqlite3 chokes trying to convert it to a
            # datetime.
            print(
                "Unable to fetch row from database, possibly due to validation error."
            )
        else:
            if row is None:
                print(f"Row {pk} not found in table {table!r}.")
                sys.exit(1)

            prettyprint_row(row)

        print()
        if not click.confirm("Are you sure you wish to delete this record?"):
            print()
            print("Operation aborted.")
            sys.exit(1)

        db.delete_by_rowid(table, pk)
        print()
        print(f"Row {pk} deleted.")


@cli.command(name="drop-column")
@click.argument("path_to_database")
@click.argument("table")
@click.argument("column")
def main_drop_column(path_to_database, table, column):
    """
    Drop a column from a table.
    """
    with Database(path_to_database) as db:
        count = db.count(table)
        print(f"WARNING: Table {table!r} contains {count} row(s) of data.")
        print()
        if not click.confirm("Are you sure you wish to drop this column?"):
            print()
            print("Operation aborted.")
            sys.exit(1)

        db.drop_column(table, column)
        print()
        print(f"Column {column!r} dropped from table {table!r}.")


@cli.command(name="get")
@click.argument("path_to_database")
@click.argument("table")
@click.argument("pk", type=int)
def main_get(path_to_database, table, pk):
    """
    Fetch a single row.
    """
    with Database(path_to_database, readonly=True) as db:
        row = db.get_by_rowid(table, pk)
        if row is None:
            print(f"Row {pk} not found in table {table!r}.")
        else:
            prettyprint_row(row)


@cli.command(name="list")
@click.argument("path_to_database")
@click.argument("table")
@click.argument("--where", default="1")
def main_list(path_to_database, table, *, where):
    """
    List the rows in the table, optionally filtered by a SQL clause.
    """
    with Database(path_to_database, readonly=True) as db:
        rows = db.list(table, where=where)

        if not rows:
            if where:
                print(f"No rows found in table {table!r} with constraint {where!r}.")
            else:
                print(f"No row founds in table {table!r}.")
        else:
            prettyprint_rows(rows)


@cli.command(name="schema")
@click.argument("path_to_database")
@click.argument("table", default="")
def main_schema(path_to_database, table=""):
    """
    Print the schema of a single table or the whole database.
    """
    with Database(path_to_database, readonly=True) as db:
        if table:
            rows = db.sql(
                "SELECT sql FROM sqlite_master WHERE name = :name AND type = 'table'",
                {"name": table},
            )

            if not rows:
                print("Table {table!r} not found.")
                sys.exit(1)

            prettyprint_row(parse_create_table_statement(rows[0]["sql"]))
        else:
            rows = db.sql(
                "SELECT name, sql FROM sqlite_master "
                + "WHERE NOT name LIKE 'sqlite%' AND type = 'table'"
            )

            if not rows:
                print("No tables found.")
                sys.exit(1)

            for i, row in enumerate(rows):
                if i != 0:
                    print()

                print(row["name"])
                prettyprint_row(parse_create_table_statement(row["sql"]))


@cli.command(name="sql")
@click.argument("path_to_database")
@click.argument("query")
@click.option("--write", is_flag=True, default=False)
def main_sql(path_to_database, query, *, write):
    """
    Run a SQL command.
    """
    readonly = not write
    with Database(path_to_database, readonly=readonly) as db:
        rows = db.sql(query)
        if rows:
            prettyprint_rows(rows)
        else:
            print("No rows found.")


@cli.command(name="update")
@click.argument("path_to_database")
@click.argument("table")
@click.argument("pk", type=int)
def main_update(path_to_database, table, pk):
    """
    Update an existing row interactively.
    """
    with Database(path_to_database) as db:
        schema_row = db.sql(
            "SELECT sql FROM sqlite_master WHERE name = :name AND type = 'table'",
            {"name": table},
            multiple=False,
        )

        if schema_row is None:
            print(f"Table {table!r} not found.")
            sys.exit(1)

        schema = parse_create_table_statement(schema_row["sql"])

        original = db.get_by_rowid(table, pk)
        if original is None:
            print(f"Row {pk} not found in table {table!r}.")

        print("To clear a value, enter NULL (case-sensitive).")
        print("To keep an existing value, leave the field blank.")
        print()

        updates = {}
        for key, value in original.items():
            if key in {"id", "created_at", "last_updated_at"}:
                continue

            definition = schema[key]
            column_type = definition.split(maxsplit=1)[0]
            print(colors.blue(key), definition)
            print("Currently:", value)
            v = input2("? ", verify=lambda v: validate_column(column_type, v))

            if v:
                if v == "NULL":
                    updates[key] = None
                else:
                    updates[key] = v

        if not updates:
            print()
            print("No updates specified.")
            sys.exit(1)

        db.update_by_rowid(table, pk, updates)
        print()
        print(f"Row {pk} updated.")


def prettyprint_rows(rows):
    for i, row in enumerate(rows):
        if i != 0:
            print()

        prettyprint_row(row)


def prettyprint_row(row):
    table = Table(padding=2)
    for key, value in row.items():
        table.row(colors.blue(key), value)
    print(table)


def parse_create_table_statement(statement):
    statement = statement.strip()
    open_index = statement.find("(")
    close_index = statement.rfind(")")
    if open_index == -1 or close_index == -1:
        raise SyntaxError

    # This crude method will fail in some cases, e.g. if a column definition contains a
    # string literal expression (e.g., in a DEFAULT clause) that contains a comma.
    columns = statement[open_index + 1 : close_index].split(",")

    r = collections.OrderedDict()
    for column in columns:
        words = column.split(maxsplit=1)
        r[words[0].strip()] = words[1].strip()

    return r


timestamp_pattern = re.compile(
    r"^[0-9]{4}-[0-9]{2}-[0-9]{2} "
    + r"[0-9]{1,2}:[0-9]{2}:[0-9]{2}\.[0-9]{6}\+[0-9]{2}:[0-9]{2}$"
)
date_pattern = re.compile(r"^[0-9]{4}-[0-9]{2}-[0-9]{2}$")


def validate_column(column_type, v):
    # Accept empty input and let the database throw an error if it's not allowed.
    #
    # (We can't also do this for type validation because SQLite doesn't do type
    #  validation.)
    if not v:
        return True

    if column_type == "TIMESTAMP":
        return timestamp_pattern.match(v)
    elif column_type == "DATE":
        return date_pattern.match(v)
    elif column_type == "INTEGER":
        return v.isdigit()
    elif column_type == "BOOLEAN":
        return v == "0" or v == "1"
    else:
        return True



def foreign_key_column(name, other_table, *, required):
    not_null = "NOT NULL" if required else ""
    return f"{name} INTEGER {not_null} REFERENCES {other_table}"


def text_column(name, *, required):
    constraint = f"CHECK({name} != '')" if required else ""
    return f"{name} TEXT NOT NULL {constraint}"
